/* 
 * Database module
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <barracuda.h>
#include <db.h>
#include <stdlib.h>
#include <stdio.h>
#include <qstring.h>
#include <keywords.h>
#include <qregexp.h>
#include <qstring.h>
#include <string.h>
#include <MessageBox.h>


/* CREATORS AND DESTRUATORS */



/* 
 * Database::Database
 * Args:    int (The initial size of the database (default = 4711))
 * Returns: undefined
 * Results: A creation of a db with initial size set
 */
Database::Database(long size)
{
    V = (Reference **) malloc (size * sizeof(Reference *));
    if (V == NULL)
    {
	fprintf(stderr, "ERROR: No memory allocking Database\n");
	exit(1);
    }

    maxLength = size;
    elements = 0;
    saveFlag = TRUE;
}

/*
 * Database::~Database
 * Args:    void
 * Returns: undefined
 * Results: Deletes itself but not the references stored
 */
Database::~Database()
{
    free(V);
}





/* MUTATORS */




/*
 * Database::sort
 * Args:    int (*cmp) (const void *, const void *) (a normal cmp)
 * Returns: void
 * Result:  The database gets sorted in nlogn using the ordering
 *          defined by cmp
 */ 
void
Database::sort(int (*fun)(Reference **, Reference **))
{
    qsort(V, elements, sizeof(Reference *), (int(*)(void*,void*))fun);
    changed();
}



int 
orderToFieldNumber(int n)
{
    if (n == 0)
	return ID;
    else return n-1+FIRST_FIELD;
}

const char *
getString(Reference *r, int field, bool expanded)
{
    if (field == ID)
	return r->id;

    if (!r->hasField(field))
    {
	fprintf(stderr, "GETSTRING: No such field: %d\n", field);
	fprintf(stderr, "I'm sorry! This is a fatal error. Aborting...\n");
	exit(1);
    }

    MacroOrString *mos = r->getField(field);
    if ((mos->type == MACRO && !expanded) || (mos->type != MACRO))
	return mos->getMacro();
    return mos->getString();
}

int
myCharCmp(int c1, int c2)
{
    if (c1 >= 'a' && c1 <= 'z')
	c1 = toupper(c1);
    if (c2 >= 'a' && c2 <= 'z')
	c2 = toupper(c2);

    if (c1 == ''  || c1 == '')
	c1 = 'U';
    if (c2 == '' || c2 == '')
	c2 = 'U';


    if (c1 == '') c1 = '[';
    if (c1 == '') c1 = 92;
    if (c1 == '') c1 = ']';
    if (c2 == '') c2 = '[';
    if (c2 == '') c2 = 92;
    if (c2 == '') c2 = ']';

    if (c1 == '') c1 = '[';
    if (c1 == '') c1 = 92;
    if (c1 == '') c1 = ']';
    if (c2 == '') c2 = '[';
    if (c2 == '') c2 = 92;
    if (c2 == '') c2 = ']';
 
    return c1-c2;
}


int
myStringCmp(const char *s1, const char *s2)
{
    int answer;
    int i;

    if (s1 == NULL || s2 == NULL)
    {
	fprintf(stderr, "MYSTRINGCMP: Sorry this is a critical error\n"
	                "Aborting...\n");
	exit(1);
    }

    for (i = 0; s1[i] && s2[i]; i++)
    {
	answer = myCharCmp(s1[i], s2[i]);
	if (answer != 0)
	    return answer;
    }

    return s1[i]-s2[i];
}

     

SortSpec *globalSortSpec;
bool expandedFlag;
int authorCompare(const char *author1, const char *author2);

int
sortCmp(Reference **rr1, Reference **rr2)
{
    Reference *r1 = *rr1;
    Reference *r2 = *rr2;
    
    for (int i = 0; i <= LAST_FIELD-FIRST_FIELD+1; i++)
    {
	int field = orderToFieldNumber(globalSortSpec->ordering[i]);
	if (field != ID)
	{
	    if (r1->hasField(field) && !r2->hasField(field))
		return 1;
	    if (!r1->hasField(field) && r2->hasField(field))
		return -1;
	    if (!r1->hasField(field) && !r2->hasField(field))
		continue;
	}
		
	const char * s1 = getString(r1, field, expandedFlag);
	const char * s2 = getString(r2, field, expandedFlag);

	if (s1 == NULL) s1 = "";
	if (s2 == NULL) s2 = "";
	
	int answer;

	if (field == AUTHOR || field == EDITOR)
	    answer = authorCompare(s1, s2);
	else
	    answer = myStringCmp(s1, s2);

	if (answer != 0)
	    return answer;
    }
    return 0;
}
	
int
sortCmpEncapsuled(Reference **rr1, Reference **rr2)
{
    if (globalSortSpec->descending)
	return -sortCmp(rr1, rr2);
    else
	return sortCmp(rr1, rr2);
}
	


void
Database::sort(SortSpec *sp, bool expanded)
{
    globalSortSpec = sp;
    expandedFlag = expanded;
    sort(sortCmpEncapsuled);
}
    




/*
 * Database::readfile
 * Args:    char *filename
 * Returns: int 
 * Results: The contents of the file gets added to the database
 */
int
Database::readFile(const char *s)
{
    bool flag = TRUE;
    if (elements > 0)
	flag = FALSE;
    int answer = readFromFile((char *) s, *this);
    saveFlag = flag;
    return answer;
}

/*
 * Database::saveFile
 * Args:    char *filename
 * Returns: int
 * Results: The contents of the database gets saved to the file
 */
int
Database::saveFile(const char *s, bool saveMacros)
{
    saveFlag = TRUE;
    return saveToFile((char *) s, *this, saveMacros);
}

/*
 * Database::addReference
 * Args:    Reference *
 * Returns: int
 * Results: A reference is added to the database. If you 
 *          want to keep a ordering you have to sort the database
 *          If failure the value returned is 0.
 */
int 
Database::addReference(Reference *r)
{
    extern int replacePolicy;
    changed();
    long collisionIndex = idCollision(r->id);
    if (collisionIndex >= 0)
    {
	switch (replacePolicy)
	{
	case SKIP_ALL:
	    delete r;
	    break;
	case REPLACE_ALL:
	    replaceRef(collisionIndex, r);
	    delete r;
	    break;
	default:
	    replacePolicy = MessageBox::showMessage("The new reference has the same key as a previously defined reference", 
						r->print(), 
						V[collisionIndex]->print());
	    switch(replacePolicy)
	    {
	    case SKIP_ALL:
		delete r;
		break;
	    case REPLACE_ALL:
		replaceRef(collisionIndex, r);
		delete r;
		break;
	    case REPLACE:
		replaceRef(collisionIndex, r);
		delete r;
		break;
	    case SKIP:
		delete r;
		break;
	    case ABORT:
		delete r;
		return 0;
		break;
	    }
	    break;
	}
    }
    else
    {
	add_Reference_(r);
    }

    return 1;
}
    

	    
	    
	    

int
Database::add_Reference_(Reference *r)
{
    changed();
    if (elements == maxLength)  /* Database full -> grow database */
    { 
	long newLength = 2*maxLength + 1;
	Reference **newV = (Reference **) malloc (newLength
						  * sizeof(Reference *));
	if (newV == NULL) 
	    return 0;
	for (long i = 0; i < elements; i++)
	    newV[i] = V[i];
	free(V);
	V = newV;
	maxLength = newLength;
    }

    V[elements++] = r;
    return 1;
}
   
void
Database::replaceRef(long index, Reference *newRef)
{
    Reference *oldRef = V[index];

    oldRef->id = newRef->id;
    oldRef->setType(newRef->getType());

    for (int i = FIRST_FIELD; i <= LAST_FIELD; i++)
    {
	if (newRef->hasField(i))
	{
	    const char *s = newRef->getField(i)->getMacro();
	    text_type_t t = newRef->getField(i)->type;
	    oldRef->setField(i, (char *) s, t, TRUE);
	}
    }
}






/*
 * Database::removeReference
 * Args:    unsigned long
 * Returns: void
 * Results: Simple version of removeReference. It moves all remaining 
 *          references so that no gaps occur. Otherwise qsort will fail.
 *          The argument is the tuple index of the reference to be removed
 *          The reference is not deleted
 */
void
Database::removeReference(long index, bool deleteRef)
{
    if (index >= elements)
    {
	fprintf(stderr, "ERROR: Index out of bounds\n");
	exit(1);
    }

    if (deleteRef)
	delete V[index];

    for (long i = index; i < elements-1; i++)
	V[i] = V[i+1];
    elements--;
    changed();
}

void
Database::removeReference(Reference *r, bool deleteRef)
{
    int index = -1;
	long i ;
    for (i = 0; i <= maxIndex(); i++) {
		if (V[i] == r)
		    index = i;
	}
    
    if (index == -1)
    {
	fprintf(stderr, "ERROR: No such reference\n");
	exit(1);
    }

    if (deleteRef)
	delete V[index];

    for (i = index; i < elements-1; i++) {
		V[i] = V[i+1];
	}
    elements--;
    changed();
}

/*
 * Database::removeAll
 * Args:    void
 * Returns: void
 * Results: Guess what!
 */
void 
Database::removeAll(bool deleteRef)
{
    if (deleteRef)
	for (long i = 0; i <= maxIndex(); i++)
	    delete V[i];
    elements = 0;
    saveFlag = TRUE;
}

/*
 * Database::merge
 * Args:    Database &b
 * Returns: int (0 on failure, 1 otherwisr)
 * Results: All references in the database b gets added to the database
 */
int
Database::merge(Database &b)
{
    changed();
    if (b.maxIndex() >= 0)
	for (int i = 0; i <= b.maxIndex(); i++)
	    if (!addReference(b[i]))
		return 0;
    return 1;
}




/* INSPECTORS */



/*
 * Database:maxIndex
 * Args:    void
 * Returns: unsigned long, which is the maximum index, or number of
 *          tuples in the database - 1.
 *          If the result eq -1 there are no tuples in the database
 * Results: void
 */
long int
Database::maxIndex()
{
    return (long int) elements-1;
}

/*
 * Database:isEmpty()
 * Args:    void
 * Returns: bool
 * Results: Guess what!
 */
bool
Database::isEmpty()
{
    return elements == 0;
}

/* 
 * Database::operator[]
 * Args:    unsigned long
 * Returns: A reference pointer at index
 * Results: void
 */ 
Reference *
Database::operator[](long index)
{
    if (index >= elements)
    {
	fprintf(stderr, "ERROR: Index out of bounds\n");
	exit(1);
    }

    return V[index];
}


/*
 * Database::select
 * Args:    int (*pred)(void *)
 * Returns: Database *
 * Results: A new database is created with all references that 
 *          make pred true. If out of memory then the returned
 *          value is NULL
 */
Database *
Database::select(int (*pred) (void *))
{
    int i;
    Database *db = new Database;
    Reference *tmp;

    if (db == NULL)
	return db;

    for (i = 0; i <= maxIndex(); i++)
	if (pred(V[i]))
	{
	    tmp = new Reference(*V[i]);
	    
	    db->addReference(tmp);
	}
    return db;
}



int
match(Reference *reg, Reference *s, bool *vect,  
      bool expanded, bool caseSense)
{
    if (!vect[s->getType() - FIRST_REFERENCE])
	    return 0;

    for (int i = FIRST_FIELD; i <= LAST_FIELD; i++)
    {	
	MacroOrString *mos = reg->getField(i);
	const char *regexp = mos->getMacro();

	QString tmp = regexp;
	tmp = tmp.stripWhiteSpace();
	if (tmp.isEmpty())
	{
	    if (s->hasField(i))
		return 0;
	}
	else
	if (!s->hasField(i))
	{
	    if (tmp != ".*")
		return 0;
	}
	else
	{
	    MacroOrString *mos2 = s->getField(i);
	    QRegExp re((const char *) tmp, caseSense);

	    if (mos2->type == STRING)
	    {
		const char *macrostr = mos2->getMacro();
		if (macrostr == NULL) macrostr = " ";
		
		if (re.match(mos2->getMacro()) == -1)
		{
		    if (tmp != ".*")
			return 0;
		}
	    }
	    else
	    {
		const char *macrostr;
		if (expanded)
		    macrostr = mos2->getString();
		else 
		    macrostr = mos2->getMacro();

		if (macrostr == NULL) macrostr = " ";
				
		if (re.match(macrostr) == -1)
		{
		    if (tmp != ".*")
			return 0;
		}
	    }
	}
    }

    QString tmp2 = reg->id;
    tmp2 = tmp2.stripWhiteSpace();    

    QString tmp3 = s->id;
    tmp3 = tmp3.stripWhiteSpace();    

    QRegExp re((const char *) tmp2, caseSense);    
    
    if (re.match((const char *) tmp3) == -1)
    	return 0;
    return 1;
}
	
	    
	    
	    
	    
	    


/*
 * Database::select
 * Args:    Reference *r, bool expanded
 * Returns: Database *
 * Results: A new database is created with all references that 
 *          matches the reference. If out of memory then the returned
 *          value is NULL. If expanded is true the regexp will be match
 *          against expanded macros, otherwise against the macro. 
 */
Database *
Database::select(Reference *r, bool vect[], bool expanded, 
		 bool caseSense)
{
    int i;
    Database *db = new Database(maxIndex()+1);

    if (db == NULL)
    {
	fprintf(stderr, "ERROR: Out of memory in Database::select\n"
		        "Aborting...\n");
	exit(1);
    }

    for (i = 0; i <= maxIndex(); i++)
	if (match(r, V[i], vect, expanded, caseSense))
	{
	    Reference *tmp = new Reference(*V[i]);
	    db->addReference(tmp);
	}

    db->changed();
    return db;
}



/*
 * Database::changed()
 * Args:    void
 * Returns: void
 * Results: Used to notify the database that something is changed
 */
void
Database::changed()
{
    saveFlag = FALSE;
}

/*
 * Database::isSaved()
 * Args:    void
 * Returns: bool
 * Results: Used to check if the database is saved
 */
bool
Database::isSaved()
{
    return saveFlag;
}


/*
 * contains
 * Args:    QString s, Reference *r, bool expanded, bool caseSense
 * Returns: bool
 * Results: return TRUE if one of the fields in r contains the string s
 */
bool
contains(QString s, Reference *r, bool expanded, bool caseSense)
{
    if ((r->id).contains(s, caseSense))
	return TRUE;
    
    for (int f = FIRST_FIELD; f <= LAST_FIELD; f++)
	if (r->hasField(f))
	{
	    QString tmp;
	    if (expanded)
		tmp = r->getField(f)->getString();
	    else
		tmp = r->getField(f)->getMacro();
	    if (tmp.contains(s, caseSense))
		return TRUE;
	}
    return FALSE;
}
	    
/*
 * Database::search
 * Args:    QString s, long startIndex, searchDir dir, bool expanded,
 *          bool caseSense
 * Returns: long (index of first occurence or -1 if not found)
 * Results: the index of the first found tuple with a field containing s is 
 *          returned 
 */
long
Database::search(QString s, long startIndex, searchDir dir, 
		 bool expanded, bool caseSense)
{
    if (startIndex < 0 || startIndex > maxIndex())
	return -1;

    if (dir == FORWARD)
    {
	for (long i = startIndex; i <= maxIndex(); i++)
	    if (contains(s, V[i], expanded, caseSense))
		return i;
    }
    else
    	for (long i = startIndex; i >= 0; i--)
	    if (contains(s, V[i], expanded, caseSense))
		return i;
    
    return -1;
}


long 
Database::idCollision(QString s)
{
    if (maxIndex() >= 0)
	for (long i = 0; i <=  maxIndex(); i++)
	    if (s == V[i]->id)
		return i;
    return -1;
}







