#include "Interface.h"
#include "Edit.h"

#define Inherited QWidget

#define NUM_SHORTCUTS      100

#define BARRACUDA_RC "brc"

Interface::Interface (QWidget* parent, const char* name)
	: Inherited (parent, name), InterfaceData (this),
	  MenuBarData (this)	
{
    extern char parserError[PARSER_ERROR_LEN+1];

    opt = new Options (NUM_SHORTCUTS);
    if (!opt->loadFromFile (BARRACUDA_RC))
	infoField->setText (parserError);
    else
	infoField->setText ("Options successfully loaded.");
    opt->print (stderr);
    
    e = new Edit();
    p = new Preferences ();
    title = new MultiLabel (this, "title");
    table = new MCListBox (this, "table");
    accel = new Shortcut (this, "accel");
    int i;
    
    QObject::connect (accel, SIGNAL (bufferChanged (const char *)),
		      infoField, SLOT (setText (const char *)));

    QObject::connect (title, SIGNAL (widthChanged (int, int)),
		      table, SLOT (setColumnWidth (int, int)));
    QObject::connect (table, SIGNAL (horizValueChanged (int)),
		      title, SLOT (setScrollOffset (int)));
    QObject::connect (table, SIGNAL (selected (int)),
		      this, SLOT (showEdit ()));


    table->setGeometry (5, 60, 390, 205);
    table->setHorizTracking (TRUE);
    for (i = 0; i < numColumns(); i++)
	table->insertColumn ();

    title->setGeometry (5, 30, 390, 25);
    title->setFrameStyle (QFrame::Panel | QFrame::Raised);
    for (i = 0; i < numColumns(); i++)
	title->insertColumn (columnLabel (i));

    setMinimumSize (225, 200);

    loadBitmaps ();

    connectMenus ();
    connectShortcuts ();

    // Init databases
    normalDb = new Database (50);
    searchRes = NULL;
    currentView = NORMALDB;
    currentName = QString ("");

    // Init options
    updatePref ();
    updateShortcuts ();
    updateFields ();
    updateTypes ();
    
    setCaption ("Barracuda - [Untitled]");
    initMacros ();
}



Interface::~Interface()
{
    delete e;
    delete p;
    delete title;
    delete table;
    delete accel;
    delete normalDb;
    if (searchRes != NULL)
	delete searchRes;
}

/* setShortcut
 *
 * Sets a shortcut and updates the menu text.
 * Args: id (as specified in MenuBarData.h)
 *       keys, n (vector with the keys)
 * =>    TRUE (if the shortcut was ok)
 *       FALSE (otherwise)
 */

bool Interface::setShortcut (int id, int keys[], int n)
{
    QPopupMenu *menu;
    QString s;
    char *t;
    int i;
    
    if (n < 0)
	return FALSE;

    if ((menu = findId (id)) == NULL)
	return FALSE;
    
    if ((t = getText (id)) == NULL)
	return FALSE;

    s = QString (t);

    if (!accel->changeItem (keys, n, id))
	return FALSE;
    
    if (n != 0)
    {
	s += "\t(" + Shortcut::keyToString(keys[0]);
	for (i = 1; i < n; i++)
	    s += " " + Shortcut::keyToString (keys[i]);
	s += ")";
    }

    menu->changeItem ((const char *)s, id);

    return TRUE;
}  

/* Resizes the window */
void Interface::resizeEvent (QResizeEvent *)
{
    int lbh = height() - leftBtn->height()  
	- table->y() - infoField->height() - 15;
    int lbw = width()-10;
    int btny = table->y() + lbh + 5;
    
    title->resize (lbw, 25);
    table->resize (lbw, lbh);

    leftBtn->move (leftBtn->x(), btny);
    leftDblArrow->move (leftDblArrow->x(), btny);
    recBtn->move (recBtn->x(), btny);
    rightBtn->move (rightBtn->x(), btny);
    rightDblArrow->move (rightDblArrow->x(), btny);

    iSrchBtn->move (lbw - iSrchBtn->width(), btny);
    rSrchBtn->move (iSrchBtn->x() - 5 - rSrchBtn->width(), btny);

    infoField->move (infoField->x(), btny + leftBtn->height() + 5);
    infoField->resize (lbw, infoField->height());
}

/* Sets the infoField when a menuItem is highlighted */
void Interface::setInfoField (int menuID)
{
    static char *menuInfoTexts[] = {
	"Begin with a new database",
	"",
	"Open a database file",
	"Open a macro file",
	"Save this database",
	"Save this database with a new name",
	"Save the macros",
	"Print this database",
	"Exit from barracuda",
	"",
	"Delete the higlighted record",
	"Edit the highlighted record",
	"Edit/View all macros",
	"Set preferences",
	"Specify what fields to view",
	"Specify keybindings",
	"Specify what reference types to view",
	"Save the options to file",
	"Search for regular expression",
	"Sort the database",
	"Start a Web browser",
	"Show exiting information about the authors",
	"", "", "", "", "", "", "", "", "", "", "", "", ""};

    infoField->setText (menuInfoTexts[menuID]);
}

/* loadBitmaps
 *
 * Args: none
 * => nothing
 * Loads the bitmaps for the buttons.
 */

void Interface::loadBitmaps ()
{
    /* Load record button pixmap */
    decorateButton (recBtn, "New", "/record.bmp");
    decorateButton (leftBtn, "<", "/back.bmp");
    decorateButton (rightBtn, ">", "/play.bmp");
    decorateButton (leftDblArrow, "<<", "/rev.bmp");
    decorateButton (rightDblArrow, ">>", "/ffw.bmp");
    decorateButton (iSrchBtn, "?>", "/rightsearch.bmp");
    decorateButton (rSrchBtn, "<?", "/leftsearch.bmp");
}


/* decorateButton
 *
 * Args: button         - the button to decorate
 *       altText        - Alternative text if the image isn't found
 *       filename       - The filename to look for the image in
 * => nothing
 */
void Interface::decorateButton (
    QPushButton *button, 
    char *altText, 
    char *filename)
{
    QPixmap *pm = new QPixmap();
    QString path = getenv("BARRACUDAHOME");
    QString def_path = BMP_PATH;
    
    if (pm->load (path + filename))
	button->setPixmap (*pm);
    else if (pm->load (def_path + filename))
	button->setPixmap (*pm);
    else
	button->setText (altText);
}

/* connectShortcuts 
 *
 * Connects the shortcuts to dummy keys, to make a keybinding
 * call accel->changeItem for that id.
 */

void Interface::connectShortcuts ()
{
    int i;
    int k[2];

    for (i = 0; i <= MAX_MENU_ID; i++)
	accel->insertItem (k, 0, i);

    accel->connectItem (EXIT_MENU_ID, this, SLOT (quit ()));
    accel->connectItem (EDIT_RECORD_MENU_ID, this, SLOT (showEdit ()));
    accel->connectItem (OPEN_DB_MENU_ID, this, SLOT (openDb()));
    accel->connectItem (SAVE_MENU_ID, this, SLOT (saveDatabase ()));
    accel->connectItem (SAVE_AS_MENU_ID, this, SLOT (saveAsDb()));
    accel->connectItem (SAVE_OPTIONS_MENU_ID, this, SLOT (saveOptions ()));
    accel->connectItem (PREFERENCES_MENU_ID, this, SLOT (pref ()));
    accel->connectItem (VIEW_TYPES_MENU_ID, this, SLOT (types ()));
}


/* connectMenus
 *
 * Connects the menu items to their slots and inserts the popup menus into
 * the menu bar, also connects the popup menus to the infoField.
 */

void Interface::connectMenus ()
{
    edit->connectItem (EDIT_RECORD_MENU_ID, this, SLOT(showEdit()));
    open->connectItem (OPEN_DB_MENU_ID, this, SLOT(openDb()));
    file->connectItem (SAVE_AS_MENU_ID, this, SLOT(saveAsDb()));
    file->connectItem (SAVE_MENU_ID, this, SLOT(saveDatabase()));
    file->connectItem (EXIT_MENU_ID, this, SLOT(quit()));
    option->connectItem (SAVE_OPTIONS_MENU_ID, this, SLOT (saveOptions ()));
    option->connectItem (PREFERENCES_MENU_ID, this, SLOT (pref ()));
    option->connectItem (VIEW_TYPES_MENU_ID, this, SLOT (types ()));


    menu->insertItem ("File", file);
    menu->insertItem ("Edit", edit);
    menu->insertItem ("Option", option);
    menu->insertItem ("Tools", tools);
    menu->insertSeparator ();
    menu->insertItem ("Help", help);

    QObject::connect (file, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (edit, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (option, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (tools, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (help, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (rec_ins, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
    QObject::connect (open, SIGNAL (highlighted (int)),
		      this, SLOT (setInfoField (int)));
}


/* numColumns
 *
 * => The number of columns that the table should include.
 */

int Interface::numColumns ()
{
    return LAST_FIELD - FIRST_FIELD + 2;
}


/* columnType
 *
 * Args: col (The column number indexed from 0)
 * => Type of that column
 */

int Interface::columnType (int col)
{
    static int ordering[] = {
	ID,
	AUTHOR,
	TITLE,
	JOURNAL,
	YEAR,
	KEY,
	VOLUME,
	NUMBER,
	MONTH,
	PAGES,
	NOTE,
	ANNOTE,
	BOOKTITLE,
	CROSSREF,
	EDITOR,
	SERIES,
	ORGANIZATION,
	PUBLISHER,
	ADDRESS,
	CHAPTER,
	TYPE,
	EDITION,
	HOWPUBLISHED,
	SCHOOL,
	INSTITUTION};
	
    if (col < 0 || col > numColumns())
	return -1;
    return ordering[col];
}

/* columnLabel
 *
 * Args: col
 * => Label for the given column
 */

char *Interface::columnLabel (int col)
{
    char *res;
    int type;

    if ((type = columnType (col)) == -1)
	return NULL;

    if ((res = strdup(keywords[type])) == NULL)
	return NULL;

    res[0] = toupper (res[0]);
    return res;
}


/* readDatabase
 *
 * Args: filename
 */

void Interface::readDatabase (const char *filename)
{
    extern char parserError[PARSER_ERROR_LEN+1];

    if (!normalDb->readFile (filename))
    {
	QMessageBox::message ("Error",  parserError, "Ok");
	infoField->setText ("");
    }
    else
    {
	setCurrentName (filename);
	infoField->setText (QString("Loaded file ")+QString (filename));
    }
}

void Interface::saveDatabase (const char *filename)
{
    extern char parserError[PARSER_ERROR_LEN+1];

    if (!normalDb->saveFile (filename))
    {
	QMessageBox::message ("Error", parserError, "Ok");
	infoField->setText ("");
    }
    else
    {
	infoField->setText (QString("Saved file ") + QString (filename));
	setCurrentName ((const char *) filename);
    }
}

void Interface::saveDatabase ()
{
    if (currentName.isEmpty())
	saveAsDb();
    else
	saveDatabase ((const char *)currentName);
}


/* discardDatabase
 *
 * Deletes the database and sees if it need saving.
 */

void Interface::discardDatabase (bool deleteRef)
{
    if (normalDb->isSaved())
	normalDb->removeAll (deleteRef);
    else if (!opt->getPreference (NO_SILLY_WARN))
    {
	if (QMessageBox::query ("Save database", 
                            "Do you want to save the database?",
                            "Yes", "No"))
	{
	    saveDatabase ();
	    normalDb->removeAll (deleteRef);
	}
    }
}
	

/* searchToNormal
 *
 * Makes the searchRes database the normal one and discards 
 * the normal database.
 */

void Interface::searchToNormal ()
{
    discardDatabase (FALSE);
    normalDb = searchRes;
    if (currentView == SEARCHRES)
	showDatabase (normalDb);
    currentView = NORMALDB;
    searchRes = NULL;
}


/* setCurrentName
 *
 * Sets the filename for the normal database.
 * Args: name
 */

void Interface::setCurrentName (const char *name)
{
    QString cap;
    currentName = QString (name);
    
    cap = QString ("Barracuda - [");
    
    if (currentName.isEmpty())
	cap += "Untitled]";
    else
	cap += currentName + "]";
    
    setCaption ((const char *)cap);
}


/* showDatabase
 *
 * Shows the given database in the table.
 * Args: db
 */

void Interface::showDatabase (Database *db)
{
    long i;

    table->setAutoUpdate (FALSE);
    table->clear (FALSE);
    for (i = 0; i <= db->maxIndex(); i++)
	showReference ((*db)[i]);
    table->setAutoUpdate (TRUE);
    table->repaint ();
}

/* showReference
 *
 * Adds a reference at the end of the table
 */

void Interface::showReference (Reference *r)
{
    MCListBoxItem *item = makeItem (r);
    
    table->insertItem (item);
}


/* makeItem
 *
 * Args: r (The reference to make an item for)
 * => The MCListBoxItem for r
 */

MCListBoxItem *Interface::makeItem (Reference *r)
{
    MCListBoxItem *item = new MCListBoxItem (FALSE);
    CHECK_PTR (item);
    int i, t;
    MacroOrString *mos;
    const char* str;
    
    item->insertString (r->id);
    for (i = 1; i < numColumns(); i++)
    {
	if ((t = columnType (i)) != -1)
	{
	    if ((mos = r->getField (t)) == NULL)
		item->insertString ("");
	    else
	    {
		if (opt->getPreference (EXPAND_MACRO))
		    str = r->getField (t)->getString ();
		else
		    str = r->getField (t)->getMacro ();

		if (str != NULL)
		    item->insertString (str);
		else
		    item->insertString ("");
	    }
	}
    }
    return item;
}


/* update*
 *
 * Updates the options
 */

void Interface::updatePref ()
{
    extern int bibdebug;
        
    bibdebug = opt->getPreference (VERBOSE_PARSE);
    title->setDoDrag (!opt->getPreference (LAZY_COLUMN));
    table->setDelimiterStyle ((enum PenStyle) opt->getPreference (DEL_STYLE));
    if (currentView == NORMALDB)
	showDatabase (normalDb);
    else
	showDatabase (searchRes);
}
    
void Interface::updateShortcuts ()
{
    shortcut **s;
    int n, i;

    if ((n = opt->getShortcuts (&s)) == 0)
	return;
    
    for (i = 0; i < n; i++)
	setShortcut (s[i]->id, s[i]->keys, s[i]->n);
}


void Interface::updateFields ()
{
    
}

void Interface::updateTypes ()
{
    int i;

    for (i = 0; i < numColumns(); i++)
	if (columnType (i) != ID)
	{
	    if (opt->getType (columnType (i)))
	    {
		table->setColumnVisible (TRUE, i);
		title->setColumnVisible (TRUE, i);
	    }
	    else
	    {
		table->setColumnVisible (FALSE, i);
		title->setColumnVisible (FALSE, i);
	    }
	}
}    

/* Show edit reference dialog */
void Interface::showEdit()
{
    int index = table->currentItem();
    Reference *r;
    MCListBoxItem *i;
    
    if (currentView == NORMALDB)
	r = (*normalDb)[index];
    else
	r = (*searchRes)[index];

    if (r != NULL)
    {
	if (e->eShow (r))
	{
	    i = makeItem (r);
	    table->changeItem (i, index);
	    normalDb->changed ();
	}
    }
}


/* openDb
 *
 * Loads a database file.
 */

void Interface::openDb()
{
    bool ok;
    
    QString filename = QFileDialog::getOpenFileName (0, "*.bib", this);
    if (!filename.isEmpty())
    {
	if (normalDb->isEmpty())
	    readDatabase (filename);
	else
	{
	    ok = QMessageBox::query ("Merge files", "Do you want to merge "
                                   "this file with the shown database?",
                                    "Yes", "No");
	    if (!ok)
	    {
		discardDatabase ();
		readDatabase (filename);
	    }
	    else
	    {
		if (currentView == NORMALDB)
		    readDatabase (filename);
		else
		{
		    searchToNormal ();
		    readDatabase (filename);
		}
	    }
	}
    }
    showDatabase (normalDb);
    currentView = NORMALDB;
}


/* saveAsDb
 *
 * Saves a database on file.
 */

void Interface::saveAsDb()
{
    QString filename = QFileDialog::getSaveFileName (0, "*.bib", this);
    if (!filename.isEmpty())
    {
	saveDatabase ((const char *) filename);
    }
}


/* quit
 *
 * Leaves the program
 */

void Interface::quit()
{
    discardDatabase ();
    qApp->quit ();
}


/* Show preference dialog */
void Interface::pref ()
{
    bool r = p->eShow (opt);

    if (r)
	updatePref ();
}

/* Show view types dialog */
void Interface::types ()
{
    if (vt->eShow (opt))
	updateTypes ();
}

/* Save options */
void Interface::saveOptions ()
{
    opt->saveToFile (BARRACUDA_RC);
}
