/* MCListBox.cc
 *
 * Implements the MCListBox widget.
 * This is a widget that has a listbox with items that
 * have a number of columns. A MultiLabel is used as a
 * title.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include "MCListBox.h"

/* MCListBoxItem::insertString
 *
 * Inserts a string into the item. If col is -1 or more than
 * the number of strings already in the item, the new string
 * is appended to the list.
 */

void MCListBoxItem::insertString (const char *text, int col)
{
    if (col < 0 || (its->count() > (uint) col))
	its->append (text);
    else
	its->insert (col, text);
}

void MCListBoxItem::removeString (int col)
{
    if (col >= 0 && ((uint) col < its->count()))
	its->remove (col);
}

const char *MCListBoxItem::lookupString (int col)
{
    if (col >= 0 && ((uint) col < its->count()))
	return its->at (col);
    else
	return NULL;
}

MCListBoxColumn::MCListBoxColumn ()
{
}

void MCListBoxColumn::setFont (QFont fn, MCListBox *parent)
{
    const QFont &oldFont = parent->font();
    parent->setFont (fn);
    setHeight (parent->fontMetrics().lineSpacing()+1);
    parent->setFont (oldFont);
    f = fn;
}



/* MCListBox
 *
 * Constructor for this class
 * Input:
 *    parent       The parent to this widget
 *    name         The name of this widget
 *    lbf          The flags that should be passed to the listbox
 *    mlf          The flags that should be passed to the multilabel
 */

MCListBox::MCListBox (QWidget *parent, const char *name,
		      WFlags f, bool autoDel)
    : QTableView (parent, name, f)
{
    doDrag = TRUE;
    doAutoScroll = TRUE;
    isTiming = FALSE;
    scrollDown = FALSE;
    curTotalHeight = 0;
    curTotalWidth = 0;
    current = -1;
    delimiter = DotLine;
    setTableFlags (Tbl_autoVScrollBar | Tbl_autoHScrollBar |
                   Tbl_smoothVScrolling | Tbl_clipCellPainting);
    switch (style()) 
    {
        case WindowsStyle:
        case MotifStyle:
            setFrameStyle (QFrame::WinPanel | QFrame::Sunken);
            setBackgroundColor (colorGroup().base());
            break;
        default:
            setFrameStyle (QFrame::Panel | QFrame::Plain);
            setLineWidth (1);
    }
    setFocusPolicy (StrongFocus);
    items = new QList<MCListBoxItem>;
    items->setAutoDelete (autoDel);
    cols = new QList<MCListBoxColumn>;
    cols->setAutoDelete (TRUE);
    doAutoDelete = autoDel;
    doLast = FALSE;
    
    QObject::connect (verticalScrollBar(), SIGNAL (valueChanged (int)),
		      SLOT (vSBChanged (int)));
    QObject::connect (horizontalScrollBar(), SIGNAL (valueChanged (int)),
		      SLOT (hSBChanged (int)));
}
    

MCListBox::~MCListBox ()
{
    delete items;
    delete cols;
}

/* insertItem
 *
 * Inserts an item into the box. An item is considered
 * to occupy one row.
 */

void MCListBox::insertItem (MCListBoxItem *i, int row)
{
    if (row < 0 || (items->count() < (uint) row))
	items->append (i);
    else
	items->insert (row, i);

    setNumRows (items->count());
    calcTotalHeight ();
}


/* insertQuickItem
 *
 * Inserts an item but does not update the number of rows or
 * calculate the total height
 */

void MCListBox::insertQuickItem (MCListBoxItem *i, int row)
{
    if (row < 0 || (items->count() < (uint) row))
	items->append (i);
    else
	items->insert (row, i);
}

/* updateRows
 *
 * Updates the number of rows and recalculates the height
 * Should be used after a number of insertQuickItem's
 */

void MCListBox::updateRows ()
{
    setNumRows (items->count());
    calcTotalHeight ();
}


/* changeItem
 *
 * Replaces the item at row with the given item.
 * If row is out of bounds nothing happen.
 * If autodel is on the previous item is deleted.
 */

void MCListBox::changeItem (MCListBoxItem *i, int row)
{
    MCListBoxItem *old;

    if (row >= 0 && (items->count() > (uint) row))
    {
	old = at (row);
	if (old != NULL)
	{
	    *(old->items()) = *(i->items());
	    calcTotalHeight ();
	    if (autoUpdate())
		repaint ();
	}
    }
}

void MCListBox::removeItem (int row)
{
    if (row >= 0 && (items->count() > (uint) row))
    {
	items->remove (row);
	setNumRows (items->count());
	calcTotalHeight ();
	if ((currentItem() != -1) && ((uint)(row) == items->count()))
	    setCurrentItem (currentItem()-1);
    }
}


/* findItem
 *
 * Returns the index of the first item that has the text in the given
 * column. -1 is returned if the text isn't found
 */

int MCListBox::findItem (const char *text, int column)
{
    QListIterator<MCListBoxItem> it(*items);
    int index = 0;

    for (; it.current(); ++it, index++)
	if (!strcmp (it.current()->lookupString (column), text))
	    return index;

    return -1;
}


/* insertColumn
 *
 * Inserts a column. The width of the column will be 10 pixels
 * until changed by a call to setColumnWidth.
 */
    
void MCListBox::insertColumn (int index)
{
    MCListBoxColumn *c = new MCListBoxColumn;
    
    c->setWidth (10);
    c->setAlignment (AlignLeft | AlignVCenter);
    c->setMargin (-1);
    c->setVisible (TRUE);
    c->setFont (font(), this);

    if (index < 0 || (cols->count() < (uint) index))
	cols->append (c);
    else
	cols->insert (index, c);
    setNumCols (cols->count());
    calcTotalWidth ();
}

 
void MCListBox::deleteColumn (int index)
{
    if (index >= 0 && (cols->count() < (uint) index))
    {
	cols->remove (index);
	setNumCols (cols->count());
	calcTotalWidth ();
    }
}

/* clear
 *
 * Clears the list, removes all items, and deletes them if
 * autoDelete is TRUE.
 */

void MCListBox::clear(bool removeCols)
{
    clearList (removeCols);
    if (autoUpdate ())
	repaint ();
}


/* count
 *
 * Returns the number of items in the MCListBox
 */

int MCListBox::count ()
{
    return items->count ();
}


/* at
 *
 * Returns a pointer to the MCListBoxItem at the given row.
 * Returns NULL if that row doesn't exist.
 */

MCListBoxItem *MCListBox::at (int row)
{
    if ((row >= 0) && ((uint) row < items->count())) 
	return items->at (row);
    else
	return NULL;
}

/* currentItem, setCurrentItem
 *
 * Sets / Returns the current item.
 * If the current item is changed a highlighted signal is emited.
 */

int MCListBox::currentItem ()
{
    return current;
}

void MCListBox::setCurrentItem (int row)
{
    int oldc;
    if (row >= 0 && ((uint) row < items->count ()))
    {
	oldc = current;
	current = row;
	if (current < topCell ())
	    setTopCell (current);
	else if (current > lastRowVisible ())
	    setTopCell (topCell() + current - lastRowVisible());
	emit highlighted (current);
	if (autoUpdate () && oldc != current)
	{
	    repaintRow (oldc);
	    repaintRow (current);
	}
    }
}

/* setColumnAlignment, setColumnMargin, setColumnVisible
 * setColumnFont, setColumnWidth
 *
 * Changes the property of a column.
 */

void MCListBox::setColumnAlignment (int alignment, int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
    {
	c->setAlignment (alignment);
	if (autoUpdate ())
	    repaintColumn (index);
    }
}

void MCListBox::setColumnMargin (int margin, int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
    {
	c->setMargin (margin);
	if (autoUpdate ())
	    repaintColumn (index);
    }
}

void MCListBox::setColumnFont (const QFont &f, int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
    {
	c->setFont (f, this);
	if (autoUpdate ())
	{
	    setAutoUpdate(FALSE);   // so the next setCellWidth won't repaint
	    setCellWidth(1);        // this forces a scrollbar update
	    setAutoUpdate(TRUE);    // so we get an update next time
	    setCellWidth(0);        // now we get the repaint event
	}
    }
}


void MCListBox::setColumnWidth (int width, int index)
{
    MCListBoxColumn *c;

    width++;
    if ((c = cols->at (index)) != NULL)
    {
	c->setWidth (width);
	calcTotalWidth ();
	if (autoUpdate ())
	{
	    setAutoUpdate(FALSE);   // so the next setCellWidth won't repaint
	    setCellWidth(1);        // this forces a scrollbar update
	    setAutoUpdate(TRUE);    // so we get an update next time
	    setCellWidth(0);        // now we get the repaint event
	}
    }
}

void MCListBox::setColumnVisible (bool visible, int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
    {
	c->setVisible (visible);
	calcTotalWidth();
	if (autoUpdate ())
	{
	    setAutoUpdate(FALSE);   // so the next setCellWidth won't repaint
	    setCellWidth(1);        // this forces a scrollbar update
	    setAutoUpdate(TRUE);    // so we get an update next time
	    setCellWidth(0);        // now we get the repaint event
	}
    }
}

/* setScrollX, setScrollY
 *
 * Sets the scroll offset of the table in the given direction
 * Args: offset
 */

void MCListBox::setScrollX (int offset)
{
    setXOffset (offset);
}

void MCListBox::setScrollY (int offset)
{
    setYOffset (offset);
}


/* columnWidth, columnAlignment,
 * columnMargin, columnFont, columnVisible
 *
 * Returns the given property for that column.
 */

int MCListBox::columnWidth (int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
	return c->width();

    return 0;
}

int MCListBox::columnAlignment (int index)
{ 
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
	return c->alignment();

    return 0;
}

int MCListBox::columnMargin (int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
	return c->margin();
    
    return 0;
}

const QFont &MCListBox::columnFont (int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
	return c->font();

    return QFont::defaultFont();
}

bool MCListBox::columnVisible (int index)
{
    MCListBoxColumn *c;

    if ((c = cols->at (index)) != NULL)
	return c->visible();

    return FALSE;
}

/* autoColumnWidth
 *
 * Sets the column width for the chosen column so that
 * all its items fit.
 * Args: index (The column width to appropriate)
 * => The width for the column
 */

int MCListBox::autoColumnWidth (int index)
{
    QListIterator<MCListBoxItem> it(*items);
    int maxWidth = 0;
    int curWidth = 0;
    bool a = autoUpdate ();
    QPainter p;
    
    setAutoUpdate (FALSE);
    p.begin (this);
    p.setFont (columnFont (index));
    for (; it.current(); ++it)
    {
	curWidth = p.fontMetrics().width (it.current()->lookupString (index));
	if (curWidth > maxWidth)
	    maxWidth = curWidth;
    }
    maxWidth += p.fontMetrics().width("x");
    p.end();
    setAutoUpdate (a);
    
    setColumnWidth (maxWidth, index);
    return maxWidth;
}


/* Various settings
 */


void MCListBox::setHorizTracking (bool en)
{
    QScrollBar *s;

    s = (QScrollBar*) horizontalScrollBar();
    s->setTracking (en);
}

void MCListBox::setVertTracking (bool en)
{
    QScrollBar *s;

    s = (QScrollBar*) verticalScrollBar();
    s->setTracking (en);
}

void MCListBox::setCutCellsV (bool en)
{
    if (en)
	setTableFlags (Tbl_cutCellsV);
    else
	clearTableFlags (Tbl_cutCellsV);
}

void MCListBox::setCutCellsH (bool en)
{
    if (en)
	setTableFlags (Tbl_cutCellsH);
    else
	clearTableFlags (Tbl_cutCellsH);
}
    
void MCListBox::setAutoScrollBar (bool enable)
{
    if (enable)
        setTableFlags (Tbl_autoVScrollBar);
    else
        clearTableFlags (Tbl_autoVScrollBar);
}

void MCListBox::setScrollBar (bool enable)
{
    if (enable)
        setTableFlags (Tbl_vScrollBar);
    else
        clearTableFlags (Tbl_vScrollBar);
}

void MCListBox::setAutoBottomScrollBar (bool enable)
{
    if (enable)
        setTableFlags (Tbl_autoHScrollBar);
    else
        clearTableFlags (Tbl_autoHScrollBar);
}

void MCListBox::setBottomScrollBar (bool enable)
{
    if (enable)
        setTableFlags (Tbl_hScrollBar);
    else
        clearTableFlags (Tbl_hScrollBar);
}

void MCListBox::setSmoothScrolling (bool enable)
{
    if (enable)
        setTableFlags (Tbl_smoothVScrolling);
    else
        clearTableFlags (Tbl_smoothVScrolling);
}

void MCListBox::setDelimiterStyle (PenStyle p)
{
    delimiter = p;
    if (autoUpdate ())
	repaint ();
}


/* Various slots
 *
 * Emit the signals generated by the scrollbars
 */

void MCListBox::vSBChanged (int value)
{
    emit vertValueChanged (value);
}

void MCListBox::hSBChanged (int value)
{
    emit horizValueChanged (value);
}


/* cellHeight, cellWidth
 *
 * Returns the height of a given row/column.
 */

int MCListBox::cellHeight (int)
{
    QListIterator<MCListBoxColumn> it(*cols);
    int max = 0;

    for (; it.current(); ++it)
	if (max < it.current()->height() && it.current()->visible())
	    max = it.current()->height();

    return max;
}

int MCListBox::cellWidth (int col)
{
    MCListBoxColumn *c = cols->at (col);
    
    if (c != NULL)
    {
	if (c->visible())
	    return c->width();
	else 
	    return 0;
    }
    else
	return 10; // Completely arbitrary
}

/* totalHeight, totalWidth
 *
 * Returns the total height / width of the box.
 */

int MCListBox::totalHeight()
{
    return curTotalHeight;
}

int MCListBox::totalWidth()
{
    return curTotalWidth;
}

/* paintCell
 *
 * Paints the given cell.
 */

void MCListBox::paintCell (QPainter *p, int row, int col)
{
    if (!columnVisible (col))
	return;

    MCListBoxItem *mci = items->at (row);
    const char *t;
    QString text;
    if (mci == NULL)
        return;
    if ((t = mci->lookupString (col)) == NULL)
	return;

    text = QString (t);
    QColorGroup g = colorGroup();
    if (current == row)
    {
        QColor   fc;                            // fill color
        if (style() == WindowsStyle)
            fc = darkBlue;                      // !!!hardcoded
        else
            fc = g.text();
//	if (hasFocus()) 
		p->fillRect (0, 0, cellWidth(col), cellHeight(row), fc);
//	    else
//		p->fillRect (1, 1, viewWidth()-2, cellHeight(row), fc);
        p->setPen (g.base());
        p->setBackgroundColor (g.text());
    } 
    else 
    {
        p->setBackgroundColor (g.base());
        p->setPen (g.text());
    }
    
    QRect cr(0, 0, cellWidth (col)-1, cellHeight (row));
    paintText (p, cr, text, col);
    if (((uint)col < cols->count()-1) || doLast)
	paintDelimiter (p, cellWidth (col)-1, 0, cellHeight (row));
    
    if (current == row /*&& hasFocus()*/)
    {
	if ( style() == WindowsStyle )
	    p->drawWinFocusRect (1, 1, cellWidth(col)-2 , cellHeight(row)-2);
	else 
	{
	    p->setPen (g.base());
	    p->setBrush (NoBrush);
	    if (col == 0)
		p->drawRect (1, 1, cellWidth(col)-2 , cellHeight(row)-2);
	    else if ((uint)col == cols->count()-1)
	    {
		    p->drawLine (cellWidth(col)-2, 1,
				 cellWidth(col)-2, cellHeight(row)-2);
		p->drawLine (0, 1, cellWidth(col)-2, 1);
		p->drawLine (0, cellHeight(row)-2, 
			     cellWidth(col)-2, cellHeight(row)-2);
	    }
	    else
	    {
		p->drawLine (0, 1, cellWidth(col), 1);
		p->drawLine (0, cellHeight(row)-2, 
			     cellWidth(col), cellHeight(row)-2);
	    }
	}
    }
    p->setBackgroundColor (g.base());
    p->setPen (g.text());
}

void MCListBox::paintText (QPainter *p, QRect r, QString &text, int col)
{
    const QFont &oldFont = p->font();
    MCListBoxColumn *c = cols->at (col);

    if (c == NULL)
	return;

    int m = c->margin();
    int align = c->alignment();

    p->setFont (c->font());
    if (m < 0)
    {
	if (r.width() > 0)
	    m = p->fontMetrics().width("x") / 2;
	else
	    m = 0;
    }
    if (align & AlignLeft)
	r.setLeft (r.left() + m);
    if (align & AlignRight)
	r.setRight (r.right() - m);
    if (align & AlignTop)
	r.setTop (r.top() + m);
    if (align & AlignBottom)
	r.setBottom (r.bottom() - m);
    p->drawText (r, align, text);
    p->setFont (oldFont);
}    

void MCListBox::paintDelimiter (
    QPainter *p, 
    int x, int y,
    int height)
{
    const QPen &oldPen = p->pen();

    p->setPen (delimiter);
    p->drawLine (x, y, x, height);
    p->setPen (oldPen);
}


void MCListBox::mousePressEvent(QMouseEvent *e)
{
    int itemClicked = findItem (e->pos().y());

    if (itemClicked != -1) 
	setCurrentItem (itemClicked);
    else if (contentsRect().contains (e->pos()) &&
	     lastRowVisible() >= (int) items->count()) 
        setCurrentItem (items->count()-1);
}

void MCListBox::mouseReleaseEvent (QMouseEvent *e)
{
    if (doDrag)
        mouseMoveEvent (e);
    if (isTiming) {
        killTimers ();
        isTiming = FALSE;
    }
}

void MCListBox::mouseDoubleClickEvent( QMouseEvent *e )
{
    mouseReleaseEvent (e);
    if (currentItem() >= 0)
        emit selected (currentItem());
}

void MCListBox::mouseMoveEvent (QMouseEvent *e)
{
    if (doDrag && (e->state() & (RightButton|LeftButton|MidButton)) != 0) 
    {
        int itemClicked = findItem (e->pos().y());
        if ( itemClicked != -1 ) 
	{
            if (isTiming) 
	    {
                killTimers ();
                isTiming = FALSE;
            }
            setCurrentItem (itemClicked);      // already current -> return
            return;
        } 
	else 
	{
            if (!doAutoScroll)
                return;
            if (e->pos().y() < frameWidth())
                scrollDown = FALSE;
            else
                scrollDown = TRUE;
            if (!isTiming) 
	    {
                isTiming = TRUE;
                startTimer( 100 );
            }
        }
    }
}


void MCListBox::keyPressEvent (QKeyEvent *e)
{
    if (numRows() == 0)
        return;
    if (currentItem() < 0)
        setCurrentItem (topCell());

    int pageSize, delta;

    switch (e->key()) 
    {
    case Key_Up:
	if (currentItem() > 0) 
	{
	    setCurrentItem (currentItem() - 1);
	    if (currentItem() < topCell())
		setTopCell (currentItem());
	}
	break;
    case Key_Down:
	if (currentItem() < (int)items->count() - 1) 
	{
	    setCurrentItem( currentItem() + 1 );
	    if (currentItem() > lastRowVisible())
		setTopCell(topCell() + currentItem() - lastRowVisible());
	}
	break;
    case Key_Next:
	delta = currentItem() - topCell();
	pageSize = lastRowVisible() - topCell();
	setTopCell (QMIN(topCell() + pageSize, (int)items->count() - 1));
	setCurrentItem(QMIN( topCell() + delta, (int)items->count() - 1 ));
	break;
    case Key_Prior:
	delta = currentItem() - topCell();
	pageSize = lastRowVisible() - topCell();
	setTopCell (QMAX (topCell() - pageSize, 0));
	setCurrentItem (QMAX (topCell() + delta, 0));
	break;

    case Key_Return:
    case Key_Enter:
	if (currentItem() >= 0)
	    emit selected(currentItem());
	break;
    default:
	break;
    }
}

void MCListBox::focusInEvent( QFocusEvent * )
{
    if (currentItem() < 0 && numRows() > 0)
        setCurrentItem (topCell());
    updateCell (currentItem(), 0);
}


void MCListBox::timerEvent (QTimerEvent *)
{
    if (scrollDown) 
    {
        if (topCell() != maxRowOffset()) 
	{
            setTopCell (topCell() + 1);
            setCurrentItem (lastRowVisible());
        }
    } 
    else 
    {
        if (topCell() != 0) 
	{
            setTopCell (topCell() - 1);
            setCurrentItem (topCell());
        }
    }
}

int MCListBox::findItem (int yPos) const
{
    return findRow (yPos);
}

bool MCListBox::itemYPos (int index, int *yPos) const
{
    return rowYPos (index, yPos);
}

/* clearList
 *
 * Deletes all items in the list.  Protected function that does NOT
 * update the list box.
 */

void MCListBox::clearList(bool removeCols)
{
    MCListBoxItem *mci;

    while (items->count()) 
    {
        mci = items->take (0);
	if (doAutoDelete)
	    delete mci;
    }

    bool a = autoUpdate();
    if (removeCols)
    {
	while (cols->count())
	    cols->take(0);
	setAutoUpdate (FALSE);
	setNumCols (0);
	setAutoUpdate (a);
    }

    if (QApplication::closingDown())
        return;
    setAutoUpdate (FALSE);
    setNumRows (0);
    current = -1;
    setTopCell (0);
    setAutoUpdate (a);
}

void MCListBox::calcTotalHeight ()
{
    uint h = 0, i;

    for (i = 0; i != items->count(); i++)
	h += cellHeight (i);

    curTotalHeight = h;
}

void MCListBox::calcTotalWidth ()
{
    int w = 0;
    int i;

    for (i = 0; (uint)i < cols->count(); i++)
	w += cellWidth (i);
    
    curTotalWidth = w;
}

void MCListBox::repaintColumn (int index)
{
    int x;

    if (autoUpdate() && colXPos (index, &x))
	repaint (x, 0, cellWidth (index), curTotalHeight);
}

void MCListBox::repaintRow (int row)
{
    int y;

    if (rowYPos (row, &y))
	repaint (0, y, curTotalWidth, cellHeight (row));
}
