/*
 * Macro/hash table module.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <barracuda.h>
#include <symtable.h>
#include <stdlib.h>
#include <stdio.h>
#include <MessageBox.h>
#include <options.h>
#include <keywords.h>

/* lookupMacro
 *
 * Args: name of macro to lookup
 * => the corresponding macro or 0 if undefined
 */
const char* lookupMacro (const char* str)
{
  string_entry_t* id;
  id_entry_t* entry;
  int lexicalLevel;		// not used

  if (!str)
    return 0;

  id = enter_identifier (str, strlen (str));

  if (!(entry = find_first (id, &lexicalLevel)))
    return 0;

  convertToCurrentFormat (entry->attr->str, entry->attr->representation);

  entry->attr->referenced = 1;

  return (const char*) entry->attr->str;
}

/* printAllMacros
 *
 * Args: file
 */
void printAllMacros (FILE* outFile)
{
  id_entry_t* ptr;
  extern Options* theOptions;

  for (ptr = sts_current_scope ()->all; ptr; ptr = ptr->next_symbol)
    {
      if (theOptions->getPreference (SAVE_UNUSED_MACROS) ||
	  ptr->attr->referenced)
	if (ptr->attr->catenated)
	  fprintf (outFile, "@string{%s = %s}\n",
		   ptr->name->ptr, (const char*) ptr->attr->str);
	else
	  fprintf (outFile, "@string{%s = \"%s\"}\n",
		   ptr->name->ptr, (const char*) ptr->attr->str);
    }
}

/* unmarkAllMacros
 *
 */
void unmarkAllMacros (void)
{
  id_entry_t* ptr;

  for (ptr = sts_current_scope ()->all; ptr; ptr = ptr->next_symbol)
    ptr->attr->referenced = 0;
}

/* printMacrosForPrint
 *
 * Args: file
 */
void printMacrosForPrint (FILE* outFile)
{
  id_entry_t* ptr;

  for (ptr = sts_current_scope ()->all; ptr; ptr = ptr->next_symbol)
    {
      QString outId("");
      QString outText("");

      /* Special treatment for reserved characters #$%&_{} */
      /* Very special for ~^\ */

      const char* origId = ptr->name->ptr;
      for (int i = 0; i < (int) strlen (ptr->name->ptr); i++)
	escapeSpecialSymbols (outId, origId[i]);

      const char* origText = (const char*) ptr->attr->str;
      for (int i = 0; i < (int) ptr->attr->str.length (); i++)
	escapeSpecialSymbols (outText, origText[i]);

      fprintf (outFile, "%s \\> %s \\\\\n",
	       (const char*) outId, (const char*) outText);
    }
}

/* enterMacro
 *
 * Args: name and text of macro to enter
 * => 0 if macro already defined and different, 1 otherwise
 */
bool enterMacro (const char* macroName, char* macroText, text_type_t t)
{
  string_entry_t* id;
  id_entry_t* entry;
  bool present;			// not used

  id = enter_identifier (macroName, strlen (macroName));

  if ((entry = find (sts_current_scope (), id)))
    {
      /* Check if different */
      entry->attr->str = texToString (entry->attr->str);

      if (strcmp ((const char*) entry->attr->str,
		  macroText))
	{
	  QString old, n;

	  old = "Key: ";
	  old += macroName;
	  old += "\nText: ";
	  old += entry->attr->str;
	  n = "Key: ";
	  n += macroName;
	  n += "\nText: ";
	  n += macroText;

	  switch (MessageBox::showMessage ("The new macro has the same key as "
					   "a previously defined macro",
					   old, n))
	    {
	    case ABORT:
	      return 0;
	    case SKIP:
	    case SKIP_ALL:
	      return 1;
	    case REPLACE:
	    case REPLACE_ALL:
	      /* Replace */
	      entry->attr->str = macroText;
	      entry->attr->representation = 0;
	      return 1;
	    };
	}
      else
	return 1;
    }

  entry = (id_entry_t*) malloc (sizeof (id_entry_t));
  entry->name = id;
  entry->attr = new attributes_t;

  /* Don't convert macros. Do that only when we use the macro. */
  entry->attr->str = macroText;
  entry->attr->representation = 0;
  entry->attr->catenated = (t == CATENATED);

  enter (sts_current_scope (), entry, &present);

  return 1;
}

/* enterPredefinedMacros
 *
 */
void enterPredefinedMacros ()
{
  /* Where can I find a list of predefined macros? */

  const struct {
    char* abbrev;
    char* text;
  } predefined[] = {
    {"jan", "January"},
    {"feb", "February"},
    {"mar", "Mars"},
    {"apr", "April"},
    {"may", "May"},
    {"jun", "June"},
    {"jul", "July"},
    {"aug", "August"},
    {"sep", "September"},
    {"oct", "October"},
    {"nov", "November"},
    {"dec", "December"},
    {(char*) 0, (char*) 0}};

  /* enter predefined macros */
  for (int i = 0; predefined[i].abbrev; i++)
    enterMacro (predefined[i].abbrev, predefined[i].text, STRING);
}

/* initMacros
 *
 */
void initMacros ()
{
  initialize_string_space ();
  open_scope ();
  enterPredefinedMacros ();
  open_scope ();
}

/* discardMacros
 *
 */
void discardMacros ()
{
  symbol_table_t* symTable = sts_current_scope ();
  id_entry_t* oldEntry;
  
  for (id_entry_t* entry = symTable->all; entry;)
    {
      delete entry->attr;
      oldEntry = entry;
      entry = entry->next_symbol;
      free (oldEntry);
    }

  /* Remove the current scope */
  close_scope ();

  free (symTable);

  /* Open new (empty) scope */
  open_scope ();
}

/* removeMacro
 *
 * Args: identifier of macro to remove
 */
void removeMacro (char* identifier)
{
  string_entry_t* id;
  id_entry_t* lastEntry = (id_entry_t*) 0,* entry;

  id = enter_identifier (identifier, strlen (identifier));

  for (entry = sts_current_scope ()->all; entry; entry = entry->next_symbol)
    {
      if (entry->name == id)
	{
	  if (!lastEntry)
	    sts_current_scope ()->all = entry->next_symbol;
	  else
	    lastEntry->next_symbol = entry->next_symbol;
	  break;
	}
      else
	lastEntry = entry;
    }
}

/* noMacros
 *
 */
bool noMacros ()
{
  return !sts_current_scope ()->all;
}

/* changeMacro
 *
 * Args: id, text
 * => 0 is bad
 */
bool changeMacro (char* macroName, char* macroText, text_type_t t)
{
  string_entry_t* id;
  id_entry_t* entry;

  id = enter_identifier (macroName, strlen (macroName));

  if ((entry = find (sts_current_scope (), id)))
    {
      entry->attr->str = macroText;
      entry->attr->representation = 1;
      entry->attr->catenated = (t == CATENATED);
      return 1;
    }

  return 0;  
}
