/* options.cc
 *
 * Implements the Options class.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <qkeycode.h>
#include <stdlib.h>

#include "options.h"
#include "MenuBarData.h"
#include "barracuda.h"
#include "Shortcut.h"

/* Options
 *
 * Creates a new options obejct with the default options.
 */

Options::Options (int max_shortcuts)
{
    this->max_shortcuts = max_shortcuts;

    if ((shortcuts = (shortcut **) 
	 malloc (sizeof (shortcut *) * max_shortcuts)) == NULL)
	exit (1);
    num_shortcuts = 0;

    initDefaults ();
    
    setDefaultIndex ();
    setDefaultPref ();
    setDefaultShortcuts ();
    setDefaultFields ();
    setDefaultTypes ();
    setDefaultColumns ();
    setDefaultFonts ();
}

/* ~Options
 *
 * Destructor. Deletes all the shortcut objects.
 */

Options::~Options ()
{
    int i;
    
    for (i = 0; i < num_shortcuts; i++)
	delete shortcuts[i];
}

/* setPreference
 *
 * Args: preference to set and its value.
 * => TRUE if the preference is valid and the value is valid.
 */

bool Options::setPreference (int pref, int value)
{
    if (pref < FIRST_PREF || pref > LAST_PREF)
	return FALSE;

    switch (pref)
    {
    case EXPAND_MACRO:
    case LAZY_COLUMN:
    case VERBOSE_PARSE:
    case NO_SILLY_WARN:
    case NO_SILLY_STARTUP:
    case CASE_SEARCH:
    case SAVE_UNUSED_MACROS:
    case CONVERT_TEX:
	if (value != 0 && value != 1)
	    return FALSE;
	break;
    case DEL_STYLE:
	if (value < 0 && value > 5)
	    return FALSE;
	break;
    }
	
    preferences[pref - FIRST_PREF] = value;
    return TRUE;
}

/* getPreference
 *
 * Args: Preference number
 * => Value of that preference
 */
int Options::getPreference (int pref)
{
    if (pref < FIRST_PREF || pref > LAST_PREF)
	return -1;

    return preferences[pref - FIRST_PREF];
}

/* setDefaultPref
 *
 * Returns all preferences to their default values.
 */
void Options::setDefaultPref()
{
    int i;

    for (i = 0; i <= (LAST_PREF - FIRST_PREF); i++)
	preferences[i] = defPreferences[i];
}

/* defaultPreference
 *
 * Args: A preference and a value
 * => TRUE if the given value is default for that preference
 */

bool Options::defaultPreference (int pref, int value)
{
    if (pref < FIRST_PREF || pref > LAST_PREF)
	return FALSE;

    return defPreferences[pref - FIRST_PREF] == value;
}

/* conflicts
 *
 * Args: Two shortcut pointers.
 * => TRUE if they conflict in any way
 */

bool Options::conflicts (shortcut *s1, shortcut *s2)
{
    int i, min;

    min = s1->n > s2->n ? s2->n : s1->n;

    for (i = 0; i < min; i++)
    {
	if (s1->keys[i] != s2->keys[i])
	    return FALSE;
    }
    
    return TRUE;
}

/* identical
 *
 * Args: Shortcut pointers
 * => TRUE if the two shortcuts are identical
 */

bool Options::identical (shortcut *s1, shortcut *s2)
{
    int i;

    if (s1->id != s2->id)
	return FALSE;

    if (s1->n != s2->n)
	return FALSE;

    for (i = 0; i < s1->n; i++)
	if (s1->keys[i] != s2->keys[i])
	    return FALSE;

    return TRUE;
}


/* sameKeys
 *
 * Args: Two shortcut pointers
 * => TRUE if they define exactly the same keys
 */

bool Options::sameKeys (shortcut *s1, shortcut *s2)
{
    if (s1->n != s2->n)
	return FALSE;

    int i;

    for (i = 0; i < s1->n; i++)
	if (s1->keys[i] != s2->keys[i])
	    return FALSE;

    return TRUE;
}


/* setShortcut
 *
 * Appends or changes the existing shortcut for that id.
 * Args: Pointer to a shortcut struct
 */

bool Options::setShortcut (shortcut *s)
{
    int i;

    for (i = 0; i < num_shortcuts; i++)
    {
	if (!identical (s, shortcuts[i]) && conflicts (s, shortcuts[i]))
	    return FALSE;
	if (shortcuts[i]->id == s->id)
	{
	    delete shortcuts[i];
	    shortcuts[i] = s;
	    return TRUE;
	}
    }
    
    if (max_shortcuts > num_shortcuts)
    {
	shortcuts[num_shortcuts++] = s;
	return TRUE;
    }
    
    barracuda_error;
    return FALSE;
}

/* setShortcut
 *
 * Changes or makes a new shortcut for the given key combination
 * Args: id    (Identifier for shortcut)
 *       keys  (Vector with keys)
 *       n     (Number of keys in keys)
 * => TRUE if ok
 */

bool Options::setShortcut (int id, int keys[], int n)
{
    shortcut *s = new shortcut;
    int i;
    
    s->id = id;
    s->n = n;
    for (i = 0; i < n; i++)
	s->keys[i] = keys[i];
    
    return (setShortcut (s));
}


/* getShortcutString
 *
 * Args: id (Identifier for shortcut)
 * => QString with the string for that id
 * NOTE: C-c is written as \C-c and so on.
 */

QString Options::getShortcutString (int id)
{
    int i, j;
    QString s = QString ("");
    int bs;

    for (i = 0; i < num_shortcuts; i++)
	if (shortcuts[i]->id == id)
	{
	    for (j = 0; j < shortcuts[i]->n; j++)
	    {
		bs = shortcuts[i]->keys[j] & (CTRL | SHIFT | ALT);
		if (bs)
		    s += "\\";
		else if (j != 0)
		    s += " ";
		s += Shortcut::keyToString (shortcuts[i]->keys[j]);
	    }
	    return s;
	}
    return s;
}


/* hasShortcut
 *
 * Args: id (Identifier for shortcut)
 * => TRUE if that id has a shortcut.
 */

bool Options::hasShortcut (int id)
{
    int i;

    for (i = 0; i < num_shortcuts; i++)
	if (shortcuts[i]->id == id)
	    return TRUE;

    return FALSE;
}


/* getFirstDefinedId
 *
 * => Identifier for first shortcut, -1 if none.
 */

int Options::getFirstDefinedId()
{
    if (num_shortcuts >= 1)
	return shortcuts[0]->id;

    return -1;
}
	

/* undefShortcut
 *
 * Undefines a keysequence.
 * Args: keys (Vector with key sequence to undef)
 *          n (Number of keys)
 * => TRUE
 */

bool Options::undefShortcut (shortcut *s)
{
    int i;

    for (i = 0; i < num_shortcuts; i++)
	if (sameKeys (shortcuts[i], s))
	    return undefShortcut (shortcuts[i]->id);
    
    return TRUE;
}

/* undefShortcut
 *
 * Removes a shortcut for a given id.
 * Args: Identifier for shortcut
 * => TRUE
 */
bool Options::undefShortcut (int id)
{
    int i, j;

    for (i = 0; i < num_shortcuts; i++)
	if (shortcuts[i]->id == id)
	{
	    delete shortcuts[i];
	    for (j = i; j < num_shortcuts-1; j++)
		shortcuts[j] = shortcuts[j+1];
	    num_shortcuts--;
	    return TRUE;
	}

    return TRUE;
}

/* getShortcuts
 *
 * => A pointer to a vector with shortcut pointers,
 *    Number of shortcuts in the vector.
 */
int Options::getShortcuts (shortcut ***s)
{
    *s = shortcuts;
    return num_shortcuts;
}

/* maxShortcuts
 *
 * => The number of shortcuts it is possible to store.
 */
int Options::maxShortcuts ()
{
    return max_shortcuts;
}

/* setDefaultShortcuts
 *
 * Returns all shortcuts to their default values
 */
void Options::setDefaultShortcuts ()
{
    int i;

    for (i = 0; i < num_shortcuts; i++)
	delete shortcuts[i];
    
    for (i = 0; i < num_defShortcuts; i++)
	shortcuts[i] = makeCopy (defShortcuts[i]);

    num_shortcuts = num_defShortcuts;
}	

/* makeCopy
 *
 * Args: shortcut *
 * => A copy of the given shortcut
 */

shortcut *Options::makeCopy (shortcut *old)
{
    shortcut *s = new shortcut;
    int i;
    CHECK_PTR (s);

    s->id = old->id;
    s->n = old->n;

    for (i = 0; i < s->n; i++)
	s->keys[i] = old->keys[i];

    return s;
}

/* defaultShortcut
 *
 * Args: A shortcut pointer
 * => TRUE if the given key sequence is default for that id.
 */

bool Options::defaultShortcut (shortcut *s)
{
    int i;

    for (i = 0; i < num_defShortcuts; i++)
	if (identical (defShortcuts[i], s))
	    return TRUE;

    return FALSE;
}


/* setField
 *
 * Args: field and how to set it
 * => TRUE if ok
 */

bool Options::setField (int field, bool enable)
{
    if (field < FIRST_REFERENCE || field > LAST_REFERENCE)
	return FALSE;

    fields[field - FIRST_REFERENCE] = enable;
    return TRUE;
}

/* getField
 *
 * Args: field
 * => How it is set
 */

bool Options::getField (int field)
{
    if (field < FIRST_REFERENCE || field > LAST_REFERENCE)
	return FALSE;

    return fields[field - FIRST_REFERENCE];
} 

/* setDefaultFields
 *
 * Sets all fields to TRUE.
 */
void Options::setDefaultFields ()
{
    int i;
    
    for (i = 0; i <= (LAST_REFERENCE - FIRST_REFERENCE); i++)
	fields[i] = defFields[i];
}

/* defaultField
 *
 * Args: A field and how it is set.
 * => TRUE if it is set to it's default value
 */

bool Options::defaultField (int field, bool enable)
{
    if (field < FIRST_REFERENCE || field > LAST_REFERENCE)
	return FALSE;

    return defFields[field - FIRST_REFERENCE] == enable;
}
	
/* setType
 *
 * Args: Type and how to set it.
 * => TRUE if ok
 */
bool Options::setType (int type, bool enable)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    types[type - FIRST_FIELD] = enable;
    return TRUE;
}

/* getType
 *
 * Args: Type
 * => Current state for type
 */
bool Options::getType (int type)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    return types[type - FIRST_FIELD];
}

/* setDefaultTypes
 *
 * Returns all types to their default values
 */
void Options::setDefaultTypes ()
{
    int i;

    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD+1); i++)
	types[i] = defTypes[i];
}

/* defaultType
 *
 * Args: Type and how is is set.
 * => TRUE if it is set to it's default value
 */

bool Options::defaultType (int type, bool enable)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    return defTypes[type - FIRST_FIELD] == enable;
}


/* setColumnWidth
 *
 * Sets the column width for the column with the given type
 */
bool Options::setColumnWidth (int type, int width)
{
    if (type == ID)
    {
	idwidth = width;
	return TRUE;
    }

    if (type < FIRST_FIELD || type > LAST_FIELD)
	return FALSE;

    if (width < 5)
	return FALSE;

    widths[type - FIRST_FIELD] = width;
    return TRUE;
}

/* getColumnWidth
 *
 * Args: type
 * => Width of the column corresponding to the given type.
 */
int Options::getColumnWidth (int type)
{
    if (type == ID)
	return idwidth;
    
    if (type < FIRST_FIELD || type > LAST_FIELD)
	return 5;

    return widths[type - FIRST_FIELD];
}

/* setDefaultColumns
 *
 * Sets all column widths to 100
 */
void Options::setDefaultColumns ()
{
    int i;

    idwidth = defIdwidth;
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD); i++)
	widths[i] = defWidths[i];
}

/* defaultWidth
 *
 * Args: A column type and it's width
 * => TRUE if it's the columns default width.
 */

bool Options::defaultWidth (int type, int width)
{
    if (type == ID)
	return defIdwidth == width;

    if (type < FIRST_FIELD || type > LAST_FIELD)
	return FALSE;

    return defWidths[type - FIRST_FIELD] == width;
}


/* setColumnFont
 *
 * Sets the column font for the column with the given type
 */
bool Options::setColumnFont (int type, QFont f)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    fonts[type - FIRST_FIELD] = f;
    return TRUE;
}

/* getColumnFont
 *
 * Args: type
 * => Font of the column corresponding to the given type.
 */
QFont Options::getColumnFont (int type)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return QFont::defaultFont();

    return fonts[type - FIRST_FIELD];
}

/* setDefaultFonts
 *
 * Sets all column fonts to the defaultfont
 */
void Options::setDefaultFonts ()
{
    int i;

    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD+1); i++)
	fonts[i] = defFonts[i];
}

/* defaultFont
 *
 * Args: A column type and it's Font
 * => TRUE if it's the columns default font.
 */

bool Options::defaultFont (int type, QFont f)
{
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    return defFonts[type - FIRST_FIELD] == f;
}

/* setColumnIndex
 *
 * Places a type at a given index.
 * Rearranges the other colunms in an appropriate way.
 * Args: column, index
 * => TRUE if ok
 */

bool Options::setColumnIndex (int type, int index)
{
    int oldi = columnIndex (type);
    int i;
    
    // Check that the values are in the allowed ranges
    if (index < 0 || index >= numColumns())
	return FALSE;
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    // Do nothing if it's already in the right position
    if (index == oldi)
	return TRUE;

    // Move the column to the front
    if (index < oldi)
    {
	for (i = oldi; i > index; i--)
	    ordering[i] = ordering[i-1];
	ordering[index] = type;
    }
    // Move the column backwards
    else
    {
	for (i = oldi; i < index; i++)
	    ordering[i] = ordering[i+1];
	ordering[index] = type;
    }

    return TRUE;
}


/* numColumns
 *
 * => Number of columns, hidden and shown
 */

int Options::numColumns ()
{
    return LAST_FIELD - FIRST_FIELD + 2;
}

/* columnType
 *
 * Args: col (The column number indexed from 0)
 * => Type of that column
 */

int Options::columnType (int index)
{
    if (index < 0 || index > numColumns ())
	return -1;
    return ordering[index];
}

/* columnIndex
 *
 * Args: type
 * => The index of the column with the given type.
 */

int Options::columnIndex (int type)
{
    int i;

    for (i = 0; i < numColumns (); i++)
	if (ordering[i] == type)
	    return i;

    return -1;
}


/* setDefaultIndex
 *
 * Returns the indicies to their default values.
 */

void Options::setDefaultIndex ()
{
    int i;

    for (i = 0; i < numColumns (); i++)
	ordering[i] = defOrdering[i];
}

/* defaultIndex
 *
 * Args: A column type and it's index
 * => TRUE if the type has it's default index.
 */

bool Options::defaultIndex (int type, int index)
{
    // Check that the values are in the allowed ranges
    if (index < 0 || index >= numColumns())
	return FALSE;
    if (type < FIRST_FIELD || type > LAST_FIELD+1)
	return FALSE;

    return defOrdering[index] == type;        
}
    
/* initDefaults
 *
 * Initializes the default options.
 */

void Options::initDefaults ()
{
    static int defP[] = {1, 1, 0, 0, 0, 1, 3, 1, 1};
    static int defT[] = {1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			 1, 0, 0, 1, 1, 0, 0, 0, 1};
    static int deforder[] = {
	ID,
	AUTHOR,
	TITLE,
	JOURNAL,
	YEAR,
	KEY,
	VOLUME,
	NUMBER,
	MONTH,
	PAGES,
	NOTE,
	ANNOTE,
	BOOKTITLE,
	CROSSREF,
	EDITOR,
	SERIES,
	ORGANIZATION,
	PUBLISHER,
	ADDRESS,
	CHAPTER,
	TYPE,
	EDITION,
	HOWPUBLISHED,
	SCHOOL,
	INSTITUTION};

    int ids[] = {OPEN_DB_MENU_ID, SAVE_MENU_ID, SAVE_AS_MENU_ID, 
		 EXIT_MENU_ID, EDIT_MACRO_MENU_ID,
		 ABOUT_MENU_ID};
    int keys[][2] = {
	{CTRL+Key_X, CTRL+Key_F},
	     {CTRL+Key_X, CTRL+Key_S},
	     {CTRL+Key_X, CTRL+Key_W},
	     {CTRL+Key_X, CTRL+Key_C},
	     {CTRL+Key_M, Key_E},
	     {CTRL+Key_H, Key_A}};
    num_defShortcuts = 6;

    int i;

    // Set default preferences
    for (i = 0; i <= (LAST_PREF - FIRST_PREF); i++)
	defPreferences[i] = defP[i];
    
    // Set default fields
    for (i = 0; i <= (LAST_REFERENCE - FIRST_REFERENCE); i++)
	defFields[i] = TRUE;

    // Set default types
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD+1); i++)
	defTypes[i] = defT[i];

    // Set default widths
    defIdwidth = 100;
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD); i++)
	defWidths[i] = 100;

    // Set default Fonts
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD+1); i++)
	defFonts[i] = QFont::defaultFont();

    // Set default ordering
    for (i = 0; i < numColumns (); i++)
	defOrdering[i] = deforder[i];

    // Set default shortcuts
    if ((defShortcuts = (shortcut **) malloc (sizeof (shortcut *) *
					      num_defShortcuts)) == NULL)
	exit (1);
    for (i = 0; i < num_defShortcuts; i++)
    {
	shortcut *s = new shortcut;
	s->id = ids[i];
	s->keys[0] = keys[i][0];
	s->keys[1] = keys[i][1];
	s->n = 2;
	defShortcuts[i] = s;
    }
}

/* assign
 *
 * Copies the values from the given object to itself.
 * Args: Options pointer
 */
void Options::assign (Options *o)
{
    int i;

    // Copy preferences
    for (i = 0; i <= (LAST_PREF - FIRST_PREF); i++)
	preferences[i] = o->preferences[i];

    // Empty my shortcut vector
    for (i = 0; i < num_shortcuts; i++)
	delete shortcuts[i];
    delete shortcuts;

    // Create new vector with correct size
    max_shortcuts = o->max_shortcuts;
    if ((shortcuts = (shortcut **) 
	 malloc (sizeof (shortcut *) * max_shortcuts)) == NULL)
	exit (1);
    num_shortcuts = 0;
    
    // Fill in vector
    for (i = 0; i < o->num_shortcuts; i++)
    {
	shortcut *s = new shortcut;
	s->id = o->shortcuts[i]->id;
	s->n = o->shortcuts[i]->n;
	for (int j = 0; j < s->n; j++)
	    s->keys[j] = o->shortcuts[i]->keys[j];

	setShortcut (s);
    }

    // Copy types
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD); i++)
	types[i] = o->types[i];

    // Copy fields
    for (i = 0; i <= (LAST_REFERENCE - FIRST_REFERENCE); i++)
	fields[i] = o->fields[i];

    // Copy columns
    idwidth = o->idwidth;
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD); i++)
	widths[i] = o->widths[i];
    
    // Copy ordering
    for (i = 0; i < numColumns (); i++)
	ordering[i] = o->ordering[i];
}    
   
    
/* loadFromFile
 *
 * Loads the options from file
 * Args: filename
 * => TRUE if ok
 */
bool Options::loadFromFile (char *filename)
{
    return loadOptionsFromFile (filename, *this);
}

/* saveToFile
 *
 * Saves options to file
 * Args: filename
 */
void Options::saveToFile (char *filename)
{
    saveOptionsToFile (filename, *this);
}


/* print
 *
 * Prints the options to a file.
 * Args: FILE *
 */
void Options::print (FILE *fp)
{
    int i;

    fprintf (fp, "Preferences:\n");
    for (i = 0; i <= LAST_PREF - FIRST_PREF; i++)
	fprintf (fp, "%d ", preferences[i]);
    fprintf (fp, "\nTypes:\n");
    for (i = 0; i <= LAST_FIELD - FIRST_FIELD; i++)
	fprintf (fp, "%d ", types[i]);
    fprintf (fp, "\nFields:\n");
    for (i = 0; i <= LAST_REFERENCE - FIRST_REFERENCE; i++)
	fprintf (fp, "%d ", fields[i]);
    fprintf (fp, "\nidWidth: %d\n", idwidth);
    fprintf (fp, "\nWidths:\n");
    for (i = 0; i <= LAST_FIELD - FIRST_FIELD; i++)
	fprintf (fp, "%d ", widths[i]);
    fprintf (fp, "\nOrdering:\n");
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD + 1); i++)
	fprintf (fp, "%d ", ordering[i]);
    fprintf (fp, "\nFonts:\n");
    for (i = 0; i <= (LAST_FIELD - FIRST_FIELD + 1); i++)
	fprintf (fp, "%s ", (const char*)externalFontRepresentation (fonts[i]));

    fprintf (fp, "\nNum shortcuts: %d\n", num_shortcuts);
    fprintf (fp, "Shortcuts:\n");
    for (i = 0; i < num_shortcuts; i++)
    {
	fprintf (fp, "  id: %d\n", shortcuts[i]->id);
	fprintf (fp, "  keys: ");
	for (int j = 0; j < shortcuts[i]->n; j++)
	    fprintf (fp, "%s ", (const char *) Shortcut::keyToString (shortcuts[i]->keys[j]));
	fprintf (fp, "\n");
    }
}
    
