/* symtable.c
 *
 * symbol table for Ada/CS compiler.
 * Magnus Christensson 1996.
 */
/* modified for Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#include <symtable.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>

/* Handwaved */
#define GLOBAL_TABLE_INITIAL_SIZE 512
#define INITIAL_STRING_SPACE_SIZE 512
#define INITIAL_SIZE 64
#define THRESHOLD 0.7

/* Module global functions and variables */

symbol_table_t* current_scope = NULL;
static inline string_entry_t* locate_identifier (const char*, int, unsigned int);
static inline unsigned int hash_function (const char*, int);
static string_entry_t* enter_string (const char*, int, int);
static string_entry_t **buckets;
static unsigned int string_table_size;
static unsigned int string_table_indices; /* string_table_size-1 */
static float string_table_threshold;
static unsigned int string_table_n_entries;
static id_entry_t* last_find_entry;
static symbol_table_t* last_find_table;
static bool last_find_jump_over = 0;

/* Open a new scope */
symbol_table_t* open_scope (void)
{
  symbol_table_t* scope;

  /* Allocate a bigger table for the global scope */
  if (current_scope == NULL)
    scope = create (GLOBAL_TABLE_INITIAL_SIZE, THRESHOLD);
  else
    scope = create (INITIAL_SIZE, THRESHOLD);
  sts_push (scope);		/* push table onto stack */
  return scope;
}

/* Close the current scope */
void close_scope (void)
{
  symbol_table_t* scope;

  scope = sts_pop ();		/* pop table */
  destroy (scope);		/* deallocate memory */
}

/* Create a new (empty) symbol table */
symbol_table_t* create (size_t size, float threshold)
{
  symbol_table_t* ret;

  ret = (symbol_table_t*) malloc (sizeof (symbol_table_t));
  ret->size = size - 1;
  ret->n_entries = 0;
  ret->threshold = threshold;
  ret->all = (id_entry_t*) 0;
  if (ret->previous)
    ret->level = ret->previous->level + 1;
  else
    ret->level = 0;
  ret->buckets = (id_entry_t**) calloc (size, sizeof (id_entry_t*));
  assert (ret->buckets != NULL); /* terminate if out of space */

  return ret;
}

/* Remove all entries in table and destroy it */
void destroy (symbol_table_t* table)
{
  free (table->buckets);
}


/* Enter entry in table.
 * Returns: the chain corresponding to entry
 * entry must be an already allocated object with the name structure filled in
 * If enter return present==TRUE then the symbol table is also set up for
 * find_next.
 */
id_entry_t* enter (symbol_table_t* table, id_entry_t* entry, bool* present)
{
  unsigned int hash_value, slot;
  id_entry_t* ptr;

  table->n_entries++;
  if ((float)table->n_entries/table->size > table->threshold)
    {
      /* Rehash */
      free (table->buckets);
      table->buckets = (id_entry_t**) calloc ((table->size+1)*2, sizeof (id_entry_t*));
      assert (table->buckets != NULL); /* terminate if out of space */
      table->size = (table->size+1)*2-1;

      /* Iterate over all symbols (the next_symbol ptr). Chains with the same
       * symbol are moved together */
      for (ptr = table->all; ptr; ptr = ptr->next_symbol)
	{
	  hash_value = ptr->name->hash_value;
	  ptr->next_chain = table->buckets[hash_value & table->size];
	  table->buckets[hash_value & table->size] = ptr;
	}
    }

  hash_value = entry->name->hash_value;
  slot = hash_value & table->size;
  for (ptr = table->buckets[slot]; ptr; ptr = ptr->next_chain)
    if (ptr->name == entry->name)
      {
	entry->next_same_name = ptr->next_same_name;
	ptr->next_same_name = entry;

	/* Set up for find_next in case of collition */
	last_find_entry = ptr;
	last_find_table = table;
	last_find_jump_over = 1;

	*present = 1;
	return entry;
      }

  /* Symbol not already in table */
  entry->next_symbol = table->all;
  table->all = entry;
  entry->next_chain = table->buckets[slot];
  entry->next_same_name = NULL;
  table->buckets[slot] = entry;
  *present = 0;
  return entry;
}

/* Returns: the first record with symbol name, or NULL if not present.
 * the lexical level of this symbol is returned through lexical_level */
id_entry_t* find_first (string_entry_t* name,
			int* lexical_level)
{
  symbol_table_t* ptr;
  id_entry_t* ret;

  for (ptr = sts_current_scope(); ptr; ptr = ptr->previous)
    if ((ret = find (ptr, name)) != NULL)
      {
	*lexical_level = ptr->level;
	last_find_entry = ret;
	last_find_table = ptr;
	last_find_jump_over = 0;
	return ret;
      }

  return NULL;
}

/* Returns: the next record with symbol name, or NULL if not present.
 * A call to find_first/find_next (with non NULL return) must precede a call to
 * find_next or the return value is undefined.
 * the lexical level of this symbol is returned through lexical_level */
id_entry_t* find_next (int* lexical_level)
{
  symbol_table_t* ptr;
  id_entry_t* ret;

  if (last_find_jump_over)
    {
      ret = last_find_entry;
      last_find_entry = last_find_entry->next_same_name;
      last_find_jump_over = 0;
      *lexical_level = last_find_table->level;
      return ret;
    }

  if ((last_find_entry->next_same_name) != NULL)
    {
      last_find_entry = last_find_entry->next_same_name;
      *lexical_level = last_find_table->level;
      return last_find_entry;
    }

  for (ptr = last_find_table->previous; ptr; ptr = ptr->previous)
    if ((ret = find (ptr, last_find_entry->name)) != NULL)
      {
	*lexical_level = ptr->level;
	last_find_entry = ret;
	last_find_table = ptr;
	return ret;
      }

  return NULL;
}


/* Returns: the entry corresponding to name, or NULL if not present */
id_entry_t* find (const symbol_table_t* table,
		  const string_entry_t* name)
{
  unsigned int slot;
  id_entry_t* ptr;

  slot = name->hash_value & table->size;

  for (ptr = table->buckets[slot]; ptr; ptr = ptr->next_chain)
    if (ptr->name == name)
      return ptr;

  return NULL;
}

/* Associate the attrs records with entry, entry and attrs must be allocated
 * by caller */
void set_attributes (id_entry_t* entry, attributes_t* attrs)
{
  entry->attr = attrs;
}

/* Get the attributes record associated with entry */
attributes_t* get_attributes (const id_entry_t* entry)
{
  return entry->attr;
}

/* Push a symbol_table onto the stack */
void sts_push (symbol_table_t* table)
{
  table->previous = current_scope; /* link to previous element in stack */
  current_scope = table;	/* update current_scope */
}

/* Pop a symbol_table of the stack */
symbol_table_t* sts_pop (void)
{
  symbol_table_t* ret;

  ret = current_scope;
  current_scope = current_scope->previous;

  return ret;
}

/* Get the top of stack */
symbol_table_t* sts_current_scope (void)
{
  return current_scope;
}

/* Returns: pointer to the entry.
 * Enters name in string table (and allocates space) */
string_entry_t* enter_identifier (const char* name, int length)
{
  return enter_string (name, length, IDENTIFIER);
}

string_entry_t* enter_string (const char* name, int length, int ret_value)
{
  unsigned int hash_value;
  string_entry_t* new_entry;

  hash_value = hash_function (name, length);

  /* If already in table, just return */
  if ((new_entry = locate_identifier (name, length, hash_value)) != NULL)
    return new_entry;
 
  string_table_n_entries++;

  if ((float)string_table_n_entries/string_table_size>string_table_threshold)
    {
      /* Rehash */
      string_entry_t **old_buckets, *ptr, *next_ptr;
      unsigned int i, old_size = string_table_size;

      old_buckets = buckets;
      string_table_size *= 2;
      string_table_indices = string_table_size - 1;
      buckets = (string_entry_t**) calloc (string_table_size, sizeof (string_entry_t*));
      assert (buckets != NULL);	/* terminate if out of space */

      /* Iterate over number of slots in old table, and over each chain */
      for (i = 0; i < old_size; i++)
	for (ptr = old_buckets[i]; ptr; ptr = next_ptr)
	  {
	    next_ptr = ptr->next;
	    ptr->next = buckets[ptr->hash_value & string_table_indices];
	    buckets[ptr->hash_value & string_table_indices] = ptr;
	  }
      free (old_buckets);
    }
  new_entry = (string_entry_t*) malloc (sizeof(string_entry_t));
  new_entry->next = buckets[hash_value & string_table_indices];
  buckets[hash_value & string_table_indices] = new_entry;
  new_entry->ptr = (char*) malloc (length + 1);
  strncpy (new_entry->ptr, name, length + 1);
  new_entry->length = length;
  new_entry->hash_value = hash_value;
  new_entry->ret_value = ret_value;

  return new_entry;
}

/* Returns: entry corresponding to identifier name, or NULL if not present */
static inline string_entry_t* locate_identifier (const char* name, int length,
						 unsigned int hash_value)
{
  string_entry_t* ptr;

  for (ptr = buckets[hash_value & string_table_indices]; ptr; ptr = ptr->next)
    if (ptr->length == (unsigned)length && 
	!strncmp (ptr->ptr, name, length))
      return ptr;
  return NULL;
}

/* Returns: dragon book hash function value of name */
static inline unsigned int hash_function (const char* name, int length)
{
  int i;
  unsigned int h=0, g;

  for (i = 0; i < length; i++)
    {
      h = (h << 4) + name[i];
      if ((g = h&0xf0000000))
	{
	  h = h ^ (g >> 24);
	  h = h ^ g;
	}
    }

  return h;
}

/* Initializes the string space */
void initialize_string_space (void)
{
  buckets = (string_entry_t**) calloc (INITIAL_STRING_SPACE_SIZE, sizeof(string_entry_t*));
  assert (buckets != NULL);	/* terminate if out of space */
  string_table_size = INITIAL_STRING_SPACE_SIZE;
  string_table_indices = INITIAL_STRING_SPACE_SIZE - 1;
  string_table_threshold = THRESHOLD;
  string_table_n_entries = 0;
}
