/* mclistbox.h
 *
 * Declares the MCListBox widget.
 * This is a widget that has a listbox with items that
 * have a number of columns.
 *
 * Barracuda 1.0.
 * Copyright (C) 1997 Magnus Christensson,
 *                    Stefan Cronert,
 *                    Stefan Hellkvist,
 *                    Magnus Ljung.
 *
 * Barracuda comes with ABSOLUTELY NO WARRANTY. For detailes see COPYING.
 */

#ifndef MCLISTBOX_H
#define MCLISTBOX_H

#include <qobject.h>
#include <qlist.h>
#include <qlistbox.h>
#include <qpen.h>
#include <qtablevw.h>
#include <qstrlist.h>
#include <qpainter.h>
#include <qkeycode.h>
#include <qscrbar.h>
#include <qapp.h>

class MCListBox;

class MCListBoxItem
{
public:
    MCListBoxItem (bool deepCopies=TRUE) 
	{ its = new QStrList (deepCopies); CHECK_PTR (its); }
    ~MCListBoxItem() { delete its; }

    void insertString (const char *text, int col=-1);
    void removeString (int col);
    const char *lookupString (int col);

    QStrList *items() { return its; }

private:
    QStrList *its;
};

class MCListBoxColumn 
{
public:
    MCListBoxColumn ();
    ~MCListBoxColumn () {}

    void setWidth (int neww)            { w = neww; }
    void setAlignment (int a)           { align = a; }
    void setMargin (int m)              { marg = m; }
    void setVisible (bool enable)       { vis = enable; }
    void setFont (QFont, MCListBox *);
    void setHeight (int h)              { hg = h; }
    
    int width ()                        { return w; }
    int alignment ()                    { return align; }
    int margin ()                       { return marg; }
    bool visible ()                     { return vis; }
    QFont& font ()                      { return f; }
    int height ()                       { return hg; }
    
private:
    int w;
    int align;
    int marg;
    bool vis;
    QFont f;
    int hg;
};


class MCListBox : public QTableView
{
    Q_OBJECT
public:
    MCListBox (QWidget *parent = 0, const char *name = 0,
	       WFlags f=0, bool autoDel=TRUE);
    ~MCListBox ();

    void insertItem (MCListBoxItem *i, int row=-1);
    void insertQuickItem (MCListBoxItem *i, int row=-1);
    void updateRows ();
    void changeItem (MCListBoxItem *i, int row);
    void removeItem (int row);
    int findItem (const char *text, int column);

    void insertColumn (int index = -1);
    void deleteColumn (int index);
    void clear (bool removeCols=FALSE);
    int count ();

    MCListBoxItem *at (int row);

    int currentItem ();
    void setCurrentItem (int row);

    int columnWidth (int index);
    int columnAlignment (int index);
    int columnMargin (int index);
    const QFont &columnFont (int index);
    bool columnVisible (int index);
    int autoColumnWidth (int index);

    bool horizTracking ()        { return horizontalScrollBar()->tracking(); }
    bool vertTracking ()         { return verticalScrollBar()->tracking(); }
    void setHorizTracking (bool en);
    void setVertTracking (bool en);
    bool drawLast ()             { return doLast; }
    void setDrawLast (bool en)   { doLast = en; }
    bool dragSelect ()           { return doDrag; }
    void setDragSelect (bool en) { doDrag = en; }
    bool autoScroll ()           { return doAutoScroll; }
    void autoScroll (bool en)    { doAutoScroll = en; }
    bool cutCellsV ()            { return testTableFlags (Tbl_cutCellsV); }
    void setCutCellsV (bool en);
    bool cutCellsH ()            { return testTableFlags (Tbl_cutCellsH); }
    void setCutCellsH (bool en);
    bool autoScrollBar ()        { return testTableFlags (Tbl_autoVScrollBar);}
    void setAutoScrollBar (bool);
    bool scrollBar ()            { return testTableFlags (Tbl_vScrollBar);}
    void setScrollBar (bool);
    bool autoBottomScrollBar ()  { return testTableFlags (Tbl_autoHScrollBar);}
    void setAutoBottomScrollBar (bool);
    bool bottomScrollBar ()      { return testTableFlags (Tbl_hScrollBar);}
    void setBottomScrollBar (bool);
    bool smoothScrolling ()      { return testTableFlags (Tbl_smoothVScrolling);}
    void setSmoothScrolling (bool);
    bool autoDelete ()           { return doAutoDelete; }
    void setAutoDelete (bool enable) { doAutoDelete = enable; items->setAutoDelete (doAutoDelete); }
    PenStyle delimiterStyle ()   { return delimiter; }
    void setDelimiterStyle (PenStyle);
    bool autoUpdate ()           { return QTableView::autoUpdate(); }
    void setAutoUpdate (bool en) { QTableView::setAutoUpdate (en); }

signals:
    void highlighted (int row);
    void selected (int row);
    void vertValueChanged (int value);
    void horizValueChanged (int value);

public slots:
    void setColumnWidth (int width, int index);
    void setColumnAlignment (int alignment, int index);
    void setColumnMargin (int margin, int index);
    void setColumnVisible (bool visible, int index);
    void setColumnFont (const QFont &f, int index);
    void setScrollX (int offset);
    void setScrollY (int offset);

private slots:
    void vSBChanged (int value);
    void hSBChanged (int value);

protected:
    int cellHeight (int);
    int cellWidth (int col);
    int totalHeight ();
    int totalWidth ();
    void paintCell (QPainter *p, int row, int col);
    void paintText (QPainter *p, QRect r, QString &text, int col);
    void paintDelimiter (QPainter *p, int x, int y, int height); 
    
    void mousePressEvent (QMouseEvent *);
    void mouseReleaseEvent (QMouseEvent *);
    void mouseDoubleClickEvent (QMouseEvent *);
    void mouseMoveEvent (QMouseEvent *);
    void keyPressEvent (QKeyEvent *);
    void focusInEvent (QFocusEvent *);
    void timerEvent (QTimerEvent *);
    void paintEvent (QPaintEvent *e) { QTableView::paintEvent(e);}
    
    int findItem (int yPos) const;
    bool itemYPos (int index, int *yPos) const;
    void clearList (bool removeCols);

private:
    void calcTotalHeight ();
    void calcTotalWidth ();
    void repaintColumn (int index);
    void repaintRow (int row);

    uint        doDrag          : 1;
    uint        doAutoScroll    : 1;
    uint        isTiming        : 1;
    uint        scrollDown      : 1;
    bool        doAutoDelete;
    bool        doLast;
    int         curTotalHeight;
    int         curTotalWidth;
    int         current;
    PenStyle    delimiter;
    QList<MCListBoxItem> *items;
    QList<MCListBoxColumn> *cols;
};

#endif
